"""CLI entrypoints."""
import asyncio
import io

import click

from payler import config, logs, process, runtime
from payler.broker import BrokerManager
from payler.db import SpoolManager


async def process_queue(loop: asyncio.events.AbstractEventLoop):
    """Start Payler using the CLI flags."""
    logger = logs.build_logger('process_queue')
    broker_url = config.get('BROKER_URL')
    mongo_url = config.get('MONGODB_URL')
    storage_manager = SpoolManager(mongo_url, loop=loop, logger=logger)
    await storage_manager.is_reachable()
    broker_manager = await BrokerManager.create(broker_url, loop=loop, logger=logger)
    broker_manager.configure(
        action=process.spool_message,
        driver=storage_manager,
    )  # action=print, driver=None
    await broker_manager.serve()


async def watch_storage(loop: asyncio.events.AbstractEventLoop):
    """Watch the storage and inject in BrokerManager exchange."""
    logger = logs.build_logger('watch_storage')
    broker_url = config.get('BROKER_URL')
    mongo_url = config.get('MONGODB_URL')
    storage_manager = SpoolManager(mongo_url, loop=loop, logger=logger)
    await storage_manager.is_reachable()
    broker_manager = await BrokerManager.create(broker_url, loop=loop, logger=logger)
    logger.info(
        'configuring SpoolManager with action=%s driver=%s',
        'process.send_message_back',
        'BrokerManager',
    )
    storage_manager.configure(
        action=process.send_message_back,
        driver=broker_manager,
    )  # action=print, driver=None
    await storage_manager.search_ready()


def listen_to_broker():
    """Watch the payload queue for payload to delay."""
    loop = asyncio.get_event_loop()
    loop.run_until_complete(process_queue(loop))
    loop.close()


def watch_payloads_ready():
    """Watch the Storage and re-inject matured payloads."""
    loop = asyncio.get_event_loop()
    loop.run_until_complete(watch_storage(loop))
    loop.close()


@click.command()
@click.option(
    '--config-file',
    help='Payler configuration file.',
    type=click.File('r'),
    required=True,
)
def run_payler(config_file: io.TextIOWrapper):
    """Register workflows and start processing."""
    logger = logs.build_logger("main")
    logger.info("Starting up payler with config_file=%s", config_file.name)
    configuration = config.load(config_file)
    loop = asyncio.get_event_loop()
    workflows = runtime.register_workflows(
        configuration['workflows'],
        loop,
    )
    logger.info("Found %d workflows", len(workflows))
    for workflow in workflows:
        workflow.register_action()
    logger.info("Firing up workflows.")
    loop.run_forever()

if __name__ == "__main__":
    listen_to_broker()



"""Configuration variables for payler."""
import os
from typing import IO

import yaml


AVAILABLE_SETTINGS = {
    'MONGODB_URL': 'mongodb://payler:secret@localhost/payler?authSource=admin',
    'BROKER_URL': 'amqp://payler:secret@localhost/payler',
}


def get(key):
    """Return the matching configuration value."""
    try:
        value = os.environ[key]
        return value
    except KeyError as err:
        if key in AVAILABLE_SETTINGS:
            return AVAILABLE_SETTINGS[key]
        raise RuntimeError(f'Missing configuration variable {err}') from err


def load(config_file: IO) -> dict:
    """Load the config file and return """
    return yaml.safe_load(config_file)



"""Database-related utilities."""
import asyncio

import logging
import typing

import motor.motor_asyncio
from motor.motor_asyncio import AsyncIOMotorCollection
import pendulum

from payler.errors import ProcessingError
from payler.structs import Payload
from payler.logs import build_logger


# NOTE: Create a common class with BrokerManager
class SpoolManager:
    """Service to store payloads and interact with the Database."""
    DEFAULT_COLLECTION_NAME = 'payloads'
    # TODO: Move to conf.py
    DEFAULT_SLEEP_DURATION = 30

    def __init__(self, url: str, loop, spool_collection: str = None, logger: logging.Logger = None):
        """Create the backend connection."""
        self.logger = logger
        if self.logger is None:
            self.logger = build_logger(self.__class__.__name__)
        self.client = motor.motor_asyncio.AsyncIOMotorClient(
            url,
            connectTimeoutMS=5000,
            serverSelectionTimeoutMS=5000,
            io_loop=loop,
        )
        self.database = self.client.get_default_database()
        self.collection_name = spool_collection or self.DEFAULT_COLLECTION_NAME
        self.collection = self.database[self.collection_name]  # type: AsyncIOMotorCollection

        self.action: typing.Callable
        self.driver = None

    def __str__(self):
        return f'{type(self)} - {self.database}'

    async def is_reachable(self) -> bool:
        """Ensure connection integrity."""
        result = await self.client.server_info()
        return result is not None

    # TODO: return object
    # NOTE: create DriverResult
    async def store_payload(self, payload: Payload, **kwargs) -> bool:
        """Store the Payload with corresponding metadatas."""
        result = await self.collection.insert_one(payload.asdict())
        self.logger.debug(
            'stored payload with id=%s reference_date=%s kwargs=%s',
            result.inserted_id,
            payload.reference_date,
            kwargs,
        )
        return result

    # TODO: Common method with BrokerManager
    def configure(self, action: typing.Callable, driver=None):
        """Configure the manager for post-spooling processing."""
        self.action = action
        self.driver = driver

    async def _search_ready(self, match_date: pendulum.datetime) -> typing.Any:
        query = {
            'reference_date': {'$lte': match_date},
        }
        documents = self.collection.find(query)
        async for doc in documents:
            yield doc
        else:
            self.logger.debug('no matching document')
        return

    async def search_ready(self, should_loop=True):
        """Find documents with a `reference_date` older than `match_date`."""
        self.logger.info(
            "Engaging database polling - Applying (action=%s, driver=%s) to events",
            self.action.__name__,
            type(self.driver),
        )
        if not should_loop:
            return await self.process_and_cleanup()
        while True:
            await self.process_and_cleanup()
            self.logger.debug('waiting for %ds', self.DEFAULT_SLEEP_DURATION)
            await asyncio.sleep(self.DEFAULT_SLEEP_DURATION)

    async def process_and_cleanup(self):
        """Find the matching jobs, process them and remove them from storage."""
        match_date = pendulum.now()
        async for doc in self._search_ready(match_date):
            try:
                result = await self.action(doc, self.driver)
                self.logger.info(
                    'Processed job with id=%s result=%s', doc['_id'], result,
                )
            except ProcessingError as err:
                self.logger.error('Could not parse %s: %s', doc['_id'], err)
                continue
            if result:
                await self.collection.delete_one({'_id': doc['_id']})
                self.logger.debug('deleted job _id=%s', doc['_id'])
        else:
            self.logger.info('Could not find any document with match_date=%s', match_date)



"""Common structures and dataclasses."""
from datetime import datetime
import json

from typing import Any
from dataclasses import dataclass, asdict


@dataclass
class Payload:
    """Wrapper around the actual payload.

    Wrap the payload content and provide metadata for payler.

    message: Original payload  # NOTE: could be pickled
    reup_time: Date after which the payload should be sent to the output queue.
    source: Queue from which the payload originated
    destination: Queue to which the payload should be sent back
    """
    message: Any
    reference_date: datetime
    source: str
    destination: str

    def asdict(self) -> dict:
        """Return a the Payload using a dictionary representation."""
        return asdict(self)

    def message_as_amqp_job(self) -> bytes:
        """Format the payload for amqp processing."""
        if isinstance(self.message, dict):
            return json.dumps(self.message).encode()
        return self.message



"""Logging-related utility functions.

Provide a friendly interface to log payler's operations.
"""
import logging


LOG_FORMAT = "%(asctime)s - [%(levelname)8s] - %(name)s: %(message)s"


# TODO: Variabilize verbosity
# TODO: Create common logger ? (or prefix with module name)
def build_logger(name: str) -> logging.Logger:
    """Configure a logger with a `name`.

    Example output:
        2020-10-30 17:55:59,927 - [    INFO ] - SpookManager: inserted document with _id=...
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter(LOG_FORMAT)
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    return logger



"""Broker-related utilities."""
import logging
import typing

import aio_pika

from payler.db import SpoolManager
from payler.errors import ProcessingError
from payler.structs import Payload
from payler.logs import build_logger


class BrokerManager:
    """Service to fetch and re-inject payloads."""
    DEFAULT_ROUTING_KEY = "payloads"
    # NOTE: Maybe bind queue to exchange + routing key in configure ?
    DEFAULT_QUEUE_NAME = "payler-jobs"

    def __init__(self):
        self.logger = None  # type: logging.Logger
        self.connection = None  # type: aio_pika.RobustConnection
        self.queue = None  # type: aio_pika.Queue
        self.action: typing.Callable
        self.driver = None  # type: SpoolManager
        self.kwargs = None

    @classmethod
    async def create(cls, url: str = None, loop = None, logger: logging.Logger = None):
        """Create the backend connection."""
        broker_manager = BrokerManager()
        broker_manager.logger = logger
        if broker_manager.logger is None:
            broker_manager.logger = build_logger(cls.__name__)
        broker_manager.connection = await aio_pika.connect_robust(
            url,
            loop=loop,
        )
        return broker_manager

    async def is_reachable(self) -> bool:
        """Ensure connection integrity."""
        result = await self.connection.ready()
        return result is None

    async def send_payload(self, payload: Payload, routing_key: str = None, **kwargs):
        """Send a Payload to self.routing_key."""
        routing = routing_key or self.DEFAULT_ROUTING_KEY
        self.logger.info(
            'Processing payload due for %s with routing_key=%s and kwargs=%s',
            payload.reference_date.isoformat(),
            routing,
            kwargs,
        )

        try:
            message = aio_pika.Message(
                body=payload.message,
            )
        except TypeError as err:
            raise ProcessingError('Invalid payload') from err
        channel = await self.connection.channel()
        result = await channel.default_exchange.publish(
            message,
            routing_key=routing,
        )
        self.logger.debug('Sent payload to %s', routing)
        await channel.close()
        return result

    def configure(self, action: typing.Callable = print, driver=None, **kwargs):
        """Configure the BrokerManager to serve the listening Queue."""
        # NOTE: could set as property / create BrokerManagerConfig
        self.action = action
        self.driver = driver
        self.kwargs = kwargs

    async def declare_queue(self, queue_name: str = None, **kwargs) -> aio_pika.Queue:
        """Instantiate a Queue and configure self.queue."""
        # NOTE: could provide default configuration and override using kwargs
        queue_name = queue_name or self.DEFAULT_QUEUE_NAME
        channel = await self.connection.channel()
        queue = await channel.declare_queue(queue_name, **kwargs)
        await channel.close()
        return queue

    async def serve(self, **kwargs):
        """Consume messages from `queue`."""
        self.logger.info(
            "Listening for events (action=%s, driver=%s)",
            self.action.__name__,
            type(self.driver),
        )
        if self.action is None:
            raise ProcessingError(
                f'Configure the {self.__class__.__name__} before using serve',
            )
        async with self.connection:
            async with self.connection.channel() as channel:
                # TODO: Variabilize queue name based on listen_queue or equivalent
                queue = await channel.declare_queue(self.DEFAULT_QUEUE_NAME, **kwargs)
                async with queue.iterator() as queue_iter:
                    async for message in queue_iter:
                        async with message.process():
                            try:
                                self.logger.debug('Processing %s', message)
                                await self.action(message, self.driver, **kwargs)
                            except ProcessingError as reason:
                                # NOTE: handle this properly (logging)
                                self.logger.error('could not parse message %s', reason)
                                continue



"""Configuration-based runtime execution.

.. code-block::yaml

    ---
    workflows:
      - name: 'Consume broker payloads and store'
        callable: "client.process_queue"
      - name: "Poll storage and re-inject in RabbitMQ"
        callable: "client.watch_storage"

The default runtime will start every workflow in a background thread.
"""
from asyncio.events import AbstractEventLoop
import asyncio
from dataclasses import dataclass, field
import importlib
from typing import Callable, Dict, List

from payler.errors import ProcessingError


@dataclass
class Workflow:
    """Atomic unit for a thread and its asyncio event_loop."""
    name: str
    action: Callable[[AbstractEventLoop], None]
    loop: AbstractEventLoop
    future: asyncio.Future = field(init=False)

    def register_action(self):
        """Register self.action as an asyncio.Future."""
        self.future = asyncio.ensure_future(self.action(self.loop))


def get_callable(location: str = "payler", process: str = None):
    """Find the action to execute using asyncio.

    `location` defaults to "payler", but anyone can provide custom
    """
    try:
        module = importlib.import_module(location)
        if process is None:
            return module
        target = module
        for attr in process.split('.'):
            target = getattr(target, attr)
        return target
    except ModuleNotFoundError as err:
        raise ProcessingError(
            f'Could not find module {location}'
        ) from err
    except AttributeError as err:
        raise ProcessingError(
            'Could not find process {process} in {location}'
        ) from err


def register_workflows(workflow_config: List[Dict[str, str]],
                       loop: AbstractEventLoop) -> List[Workflow]:
    """Transform the `workflows` config entry in a list of `Workflow`.

    This function takes the `.workflows` list and creates the corresponding
    Workflow, which can then be used to start background processes.

    Example input:

    .. code-bloc::yaml

        - name: "Consume broker payloads and store in MongoDB"
          callable: "client.process_queue"
        - name: "Poll storage and re-inject in RabbitMQ"
          callable: "client.watch_storage"
    """
    workflows = []
    for item in workflow_config:
        workflow = Workflow(
            item.get('name', 'unnamed'),
            get_callable(process=item.get('callable')),
            loop,
        )
        workflows.append(workflow)
    return workflows



"""Processing functions for broker operations."""
import aio_pika
import pendulum

from payler.broker import BrokerManager
from payler.db import SpoolManager
from payler.structs import Payload


async def spool_message(message: aio_pika.Message, driver: SpoolManager, **kwargs):
    """Decode and spool `message` using `driver`."""
    delay = int(message.headers.get('x-delay'))
    # NOTE: transform default destination in constant
    destination = message.headers.get('x-destination', 'payler-out')
    data = message.body
    now = pendulum.now()
    reference = now.add(microseconds=delay * 1000)  # switch form us to ms
    # TODO: Variabilize source
    source = 'payler-jobs'
    payload = Payload(
        data,
        reference,
        source,
        destination,
    )
    result = await driver.store_payload(payload, **kwargs)
    # TODO: do correct post-processing logging
    return result, payload


async def send_message_back(document: dict, driver: BrokerManager, **kwargs):
    """Inject the Payload back in the Broker."""
    payload = Payload(
        message=document.get('message'),
        reference_date=document.get('reference_date'),
        source=document.get('source'),
        destination=document.get('destination'),
    )
    return await driver.send_payload(payload, routing_key=payload.destination, **kwargs)



"""Packaging file for payler."""
from distutils.core import setup


__version__ = "0.1.1"
URL = "https://github.com/tbobm/payler/archive/{}.tar.gz".format(__version__)

setup(
    name="payler",
    packages=["payler"],
    install_requires=[
        "aio_pika",
        "click",
        "motor",
        "pendulum",
        "pymongo",
        "PyYAML",
    ],
    version=__version__,
    description="Broker payload spooler",
    author="Theo 'Bob' Massard",
    author_email="tbobm@protonmail.com",
    url="https://github.com/tbobm/payler",
    download_url=URL,
    classifiers=[
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    entry_points={
        "console_scripts": [
            "payler=payler.client:run_payler",
        ],
    },
)
